/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.InstanceNotFoundException;
import org.ow2.bonita.facade.exception.TaskNotFoundException;
import org.ow2.bonita.facade.exception.VariableNotFoundException;
import org.ow2.bonita.facade.internal.InternalQueryRuntimeAPI;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityFullInstance;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.TaskState;
import org.ow2.bonita.facade.runtime.impl.ActivityInstanceImpl;
import org.ow2.bonita.facade.runtime.impl.ProcessInstanceImpl;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.util.EngineEnvTool;
import org.ow2.novabpm.util.Misc;


/**
 * @author Pierre Vigneras
 */
public class QueryRuntimeAPIImpl implements InternalQueryRuntimeAPI {

  /*
  public QueryRuntimeAPIImpl(String CONF_NAME) {
    this.confName = CONF_NAME;
  }
   */
  protected QueryRuntimeAPIImpl() {
  }

  public ProcessInstance getProcessInstance(final ProcessInstanceUUID instanceUUID, 
      final String queryList) throws InstanceNotFoundException {
    FacadeUtil.checkArgsNotNull(instanceUUID);
    final ProcessInstance result = EngineEnvTool.getAllQueriers(queryList).getProcessInstance(instanceUUID);
    if (result == null) {
      throw new InstanceNotFoundException(instanceUUID);
    }
    return new ProcessInstanceImpl(result);
  }

  public Set<ProcessInstance> getProcessInstances(final String queryList) {
    final Set<ProcessInstance> result = new HashSet<ProcessInstance>();
    for (final ProcessInstance record : EngineEnvTool.getAllQueriers(queryList).getProcessInstances()) {
      result.add(new ProcessInstanceImpl(record));
    }
    return result;
  }

  public Set<ProcessInstance> getProcessInstances(final ProcessDefinitionUUID processUUID, final String queryList) {
    final Set<ProcessInstance> result = new HashSet<ProcessInstance>();
    for (final ProcessInstance record : EngineEnvTool.getAllQueriers(queryList).getProcessInstances(processUUID)) {
      result.add(new ProcessInstanceImpl(record));
    }
    return result;
  }

  public ActivityInstance<ActivityBody> getActivityInstance(final ProcessInstanceUUID instanceUUID, final String activityId,
      final String iterationId, final String activityInstanceId, 
      final String queryList) throws ActivityNotFoundException, InstanceNotFoundException {
    final ActivityInstance<ActivityBody> result =
      EngineEnvTool.getAllQueriers(queryList).getActivityInstance(instanceUUID, activityId, iterationId, activityInstanceId);
    if (result == null) {
      if (EngineEnvTool.getAllQueriers(queryList).getProcessInstance(instanceUUID) == null) {
        throw new InstanceNotFoundException(instanceUUID);
      }
      throw new ActivityNotFoundException(instanceUUID, activityId, iterationId);
    }
    return new ActivityInstanceImpl<ActivityBody>(result);
  }

  public Set<ActivityInstance<ActivityBody>> getActivityInstances(final ProcessInstanceUUID instanceUUID, 
      final String queryList) {
    final Set<ActivityInstance<ActivityBody>> result = new HashSet<ActivityInstance<ActivityBody>>();
    for (final ActivityInstance<ActivityBody> record 
        : EngineEnvTool.getAllQueriers(queryList).getActivityInstances(instanceUUID)) {
      result.add(new ActivityInstanceImpl<ActivityBody>(record));
    }
    return result;
  }

  public Set<ActivityInstance<ActivityBody>> getActivityInstances(final ProcessInstanceUUID instanceUUID,
      final String activityId, final String queryList) throws ActivityNotFoundException {
    final Set<ActivityInstance<ActivityBody>> result = new HashSet<ActivityInstance<ActivityBody>>();
    for (final ActivityInstance<ActivityBody> record 
        : EngineEnvTool.getAllQueriers(queryList).getActivityInstances(instanceUUID)) {
      if (record.getActivityId().equals(activityId)) {
        result.add(new ActivityInstanceImpl<ActivityBody>(record));
      }
    }
    if (result.isEmpty()) {
      throw new ActivityNotFoundException(instanceUUID, activityId);
    }
    return result;
  }

  public Set<ActivityInstance< TaskInstance >> getTasks(final ProcessInstanceUUID instanceUUID, final String queryList) {
    final Set<ActivityInstance< TaskInstance >> result = new HashSet<ActivityInstance< TaskInstance >>();
    for (final ActivityInstance< TaskInstance > record : EngineEnvTool.getAllQueriers(queryList).getTaskInstances(instanceUUID)) {
      result.add(new ActivityInstanceImpl< TaskInstance >(record));
    }
    return result;
  }

  public Collection<ActivityInstance< TaskInstance >> getTaskList(final ProcessInstanceUUID instanceUUID,
      final TaskState taskState, final String queryList) throws InstanceNotFoundException {
    return getTaskListUser(instanceUUID, EngineEnvTool.getUserId(), taskState, queryList);
  }

  public Collection<ActivityInstance< TaskInstance >> getTaskList(final TaskState taskState, final String queryList) {
    return getTaskListUser(EngineEnvTool.getUserId(), taskState, queryList);
  }

  public ActivityInstance< TaskInstance > getTask(final TaskUUID taskUUID, final String queryList) throws TaskNotFoundException {
    final ActivityInstance< TaskInstance > taskInstance = EngineEnvTool.getAllQueriers(queryList).getTaskInstance(taskUUID);
    if (taskInstance == null) {
      throw new TaskNotFoundException(taskUUID);
    }
    return new ActivityInstanceImpl< TaskInstance >(taskInstance);
  }

  public Map<String, Object> getActivityInstanceVariables(final ProcessInstanceUUID instanceUUID,
      final String activityId, final String iterationId, final String activityInstanceId, final String queryList)
    throws ActivityNotFoundException, InstanceNotFoundException {

    final ActivityInstance<ActivityBody> activityInst =
      getActivityInstance(instanceUUID, activityId, iterationId, activityInstanceId, queryList);
    if (activityInst == null) {
      throw new ActivityNotFoundException(instanceUUID, activityId);
    }
    return activityInst.getLastKnownVariableValues();
  }

  public Object getActivityInstanceVariable(final ProcessInstanceUUID instanceUUID,
      final String activityId, final String iterationId,
      final String activityInstanceId, final String variableId, final String queryList) throws InstanceNotFoundException,
      ActivityNotFoundException, VariableNotFoundException {
    final Map<String, Object> variables = 
      getActivityInstanceVariables(instanceUUID, activityId, iterationId, activityInstanceId, queryList);
    if (variables == null || !variables.containsKey(variableId)) {
      throw new VariableNotFoundException(instanceUUID, activityId, variableId);
    }
    return variables.get(variableId);
  }

  public Map<String, Object> getProcessInstanceVariables(
      final ProcessInstanceUUID instanceUUID, final String queryList) throws InstanceNotFoundException {
    final ProcessInstance processInstance = getProcessInstance(instanceUUID, queryList);
    if (processInstance == null) {
      throw new InstanceNotFoundException(instanceUUID);
    }
    return processInstance.getLastKnownVariableValues();
  }

  public Object getProcessInstanceVariable(final ProcessInstanceUUID instanceUUID,
      final String variableId, final String queryList) throws InstanceNotFoundException, VariableNotFoundException {
    final Map<String, Object> variables = getProcessInstanceVariables(instanceUUID, queryList);
    if (variables == null || !variables.containsKey(variableId)) {
      throw new VariableNotFoundException(instanceUUID, variableId);
    }
    return variables.get(variableId);
  }

/*
  private Map<String, Object> getVariablesInitialValue(List<VariableDefinitionImpl> variableDefinitions) {
    final Map<String, Object> values = new HashMap<String, Object>();
    if (variableDefinitions != null) {
      for (VariableDefinitionImpl variableDefinition : variableDefinitions) {
        final String variableId = variableDefinition.getKey();
        Object initialValue = null;
        if (variableDefinition.getSourceDescriptor() != null) {
          initialValue = variableDefinition.getSourceDescriptor().cget();
        }
        values.put(variableId, initialValue);
      }
    }
    return values;
  }

  @SuppressWarnings("unchecked")
  public Map<String, Object> getActivityVariableDefinitions(ProcessDefinitionUUID processId,
      String activityId) throws ProcessNotFoundException, ActivityNotFoundException {
    final XpdlProcess process  = EnvTool.getRepository().findXpdlProcessById(processId);
    if (process == null) {
      throw new ProcessNotFoundException(processId);
    }
    final NodeImpl node = process.getNode(activityId);
    if (node == null) {
      throw new ActivityNotFoundException(processId, activityId);
    }
    final AbstractActivity activity = (AbstractActivity) node.getBehaviour();
    final List<VariableDefinition> variableDefinitions = new ArrayList<VariableDefinition>();
    if (activity.getVariableDefinitions() != null) {
      variableDefinitions.addAll(activity.getVariableDefinitions());
    }
    if (process.getVariableDefinitions() != null) {
      variableDefinitions.addAll(process.getVariableDefinitions());
    }
    return getVariablesInitialValue((List) variableDefinitions);
  }

  @SuppressWarnings("unchecked")
  public Map<String, Object> getProcessVariableDefinitions(ProcessDefinitionUUID processId) throws ProcessNotFoundException {
    final XpdlProcess process  = EnvTool.getRepository().findXpdlProcessById(processId);
    if (process == null) {
      throw new ProcessNotFoundException(processId);
    }
    final List<VariableDefinition> variableDefinitions = process.getVariableDefinitions();
    return getVariablesInitialValue((List) variableDefinitions);
  }

*/

  public ActivityInstance<ActivityBody> getActivityInstance(
      final ActivityInstanceUUID activityUUID, final String queryList) throws ActivityNotFoundException {
    final ActivityInstance<ActivityBody> activity = EngineEnvTool.getAllQueriers(queryList).getActivityInstance(activityUUID);
    if (activity == null) {
      throw new ActivityNotFoundException(activityUUID);
    }
    return new ActivityInstanceImpl<ActivityBody>(activity);
  }

  public Object getActivityInstanceVariable(final ActivityInstanceUUID activityUUID,
      final String variableId, final String queryList) throws ActivityNotFoundException,
      VariableNotFoundException {
    final ActivityInstance<ActivityBody> activity = getActivityInstance(activityUUID, queryList);
    final Map<String, Object> variableValues = activity.getLastKnownVariableValues();

    if (!variableValues.containsKey(variableId)) {
      throw new VariableNotFoundException(activityUUID, variableId);
    }
    return activity.getLastKnownVariableValues().get(variableId);
  }

  public Map<String, Object> getActivityInstanceVariables(
      final ActivityInstanceUUID activityUUID, final String queryList) throws ActivityNotFoundException {
    final ActivityInstance<ActivityBody> activity = getActivityInstance(activityUUID, queryList);
    return activity.getLastKnownVariableValues();
  }

  public Collection<ActivityInstance<TaskInstance>> getTaskList(final ProcessInstanceUUID instanceUUID,
      final String userId, final TaskState taskState, final String queryList) throws InstanceNotFoundException {
    return getTaskListUser(instanceUUID, userId, taskState, queryList);
  }

  public Collection<ActivityInstance<TaskInstance>> getTaskList(
      final String userId, final TaskState taskState, final String queryList) {
    return getTaskListUser(userId, taskState, queryList);
  }

  private Collection<ActivityInstance<TaskInstance>> getTaskListUser(final ProcessInstanceUUID instanceUUID, final String userId,
      final TaskState taskState, final String queryList) throws InstanceNotFoundException {
    FacadeUtil.checkArgsNotNull(instanceUUID, taskState, userId);
    final ProcessInstance processInstance = getProcessInstance(instanceUUID, queryList);
    if (processInstance == null) {
      throw new InstanceNotFoundException(instanceUUID);
    }
    final Querier journal = EngineEnvTool.getAllQueriers(queryList);
    final Set<ActivityFullInstance<TaskInstance>> userInstanceTasks =
      journal.getUserInstanceTasks(userId, instanceUUID, taskState);

    final Collection<ActivityInstance<TaskInstance>> todos = new ArrayList<ActivityInstance<TaskInstance>>();
    for (final ActivityInstance<TaskInstance> taskActivity : userInstanceTasks) {
      todos.add(new ActivityInstanceImpl<TaskInstance>(taskActivity));
    }
    return todos;
  }
  private Collection<ActivityInstance< TaskInstance >> getTaskListUser(
      final String userId, final TaskState taskState, final String queryList) {
    FacadeUtil.checkArgsNotNull(userId, taskState);
    final Collection<ActivityInstance< TaskInstance >> todos = new ArrayList<ActivityInstance< TaskInstance >>();
    final Querier querier = EngineEnvTool.getAllQueriers(queryList);
    final Set<ActivityFullInstance<TaskInstance>> taskActivities = querier.getUserTasks(userId, taskState);
    for (final ActivityFullInstance<TaskInstance> taskActivity : taskActivities) {
      todos.add(new ActivityInstanceImpl<TaskInstance>(taskActivity));
    }
    return todos;
  }


  public Object getVariable(final ActivityInstanceUUID activityUUID, final String variableId, final String queryList)
    throws ActivityNotFoundException, VariableNotFoundException {
    try {
      return getActivityInstanceVariable(activityUUID, variableId, queryList);
    } catch (final VariableNotFoundException e) {
      final ActivityFullInstance<ActivityBody> activity = 
        EngineEnvTool.getAllQueriers(queryList).getActivityInstance(activityUUID);
      if (activity == null) {
        throw new ActivityNotFoundException(activityUUID);
      }
      try {
        return getProcessInstanceVariable(activity.getProcessInstanceUUID(), variableId, queryList);
      } catch (final InstanceNotFoundException e1) {
        // If activity exists, the process instance must exist too.
        Misc.unreachableStatement();
        return null;
      }
    }
  }

  public Map<String, Object> getVariables(final ActivityInstanceUUID activityUUID, 
      final String queryList) throws ActivityNotFoundException {
    final ActivityFullInstance<ActivityBody> activity = EngineEnvTool.getAllQueriers(queryList).getActivityInstance(activityUUID);
    if (activity == null) {
      throw new ActivityNotFoundException(activityUUID);
    }
    try {
      final Map<String, Object> allVariables = new HashMap<String, Object>();
      final Map<String, Object> localVariables = activity.getLastKnownVariableValues();
      final Map<String, Object> globalVariables = getProcessInstanceVariables(activity.getProcessInstanceUUID(), queryList);
      //add global first because if some variables are in both local and global
      //we want to keep local value
      allVariables.putAll(globalVariables);
      allVariables.putAll(localVariables);
      return allVariables;
    } catch (final InstanceNotFoundException e) {
      // If activity exists, the process instance must exist too.
      Misc.unreachableStatement();
      return null;
    }
  }
}
